package org.freecompany.util.services;

import java.util.HashMap;
import java.util.Map;
import java.util.logging.Logger;

import static java.util.Collections.synchronizedMap;
import static java.util.logging.Level.FINE;
import static java.util.logging.Logger.getLogger;

/**
 * The discovery class provides factory methods for creating instances of
 * service information objects and a caching mechanism. All methods of this
 * class are thread safe. 
 * 
 * <p>These service objects can be used to find concrete implementations of
 * service provider interfaces. The service discovery factory also provides
 * caching for these information objects.</p>
 * 
 * <p>For example, to find the first known concrete provider for the interface {@code ServiceInterface}:</p>
 *
 * <p>{@code ServiceInterface service = ServiceDiscovery.getInstance().findService( ServiceInterface.class);}</p>
 */
public class ServiceDiscovery {
	
	private final static Logger logger = getLogger( ServiceDiscovery.class.getName());

	protected static Map< String, ServiceInfo> cache;

	/**
	 * Enables {@link ServiceInfo} caching within this helper. With caching enabled,
	 * {@link ServiceInfo} instances are cached for each system identifier to eliminate
	 * the need to reparse configuration properties.
	 * <p/>
	 * Caching is disabled by default.
	 */
	public static void enableCaching() {
		if ( cache == null) cache = synchronizedMap( new HashMap< String, ServiceInfo>());
	}

	/**
	 * Disables {@link ServiceInfo} caching within this helper. With caching disabled,
	 * a new {@link ServiceInfo} instance is created for each {@link #getInstance()} call,
	 * which may require reparsing properties files.
	 * <p/>
	 * Caching is disabled by default.
	 */
	public static void disableCaching() {
		cache = null;
	}

	/**
	 * Returns a {@link ServiceInfo} instance that can access service providers
	 * specified as a system property or through the class path using the standard
	 * JAR services mechanism.
	 *
	 * @return a {@link ServiceInfo} object that can be used to locate service providers.
	 */
	public static ServiceInfo getInstance() {
		return getInstance( null);
	}

	/**
	 * Returns a {@link ServiceInfo} instance that can access service providers
	 * set as a system wide preference, specified as a system property, or through
	 * the class path using the standard JAR services mechanism.
	 *
	 * @return a {@link ServiceInfo} object that can be used to locate service providers.
	 */
	public static ServiceInfo getInstance( final String system) {
		if ( cache != null && cache.containsKey( system)) {
			logger.log( FINE, "Using cache to resolve ''{0}''.", system);
			return cache.get( system);
		}

		final ServiceInfo info = new ServiceInfo( system); 
		if ( cache != null) cache.put( system, info);
		return info;
	}
}
