package org.freecompany.util.services;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.net.URL;
import java.security.PrivilegedAction;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.LinkedList;
import java.util.Properties;
import java.util.logging.Logger;
import java.util.regex.Pattern;

import static java.lang.Class.forName;
import static java.security.AccessController.doPrivileged;
import static java.util.Collections.list;
import static java.util.logging.Level.FINE;
import static java.util.logging.Level.WARNING;
import static java.util.logging.Logger.getLogger;
import static java.util.regex.Pattern.compile;

/**
 * A service information object provides an interface that can be used to query
 * for provider instances defined in system properties, system wide JVM settings, or
 * service descriptors in the class path.
 */
public class ServiceInfo {
	
	private final static Logger logger = getLogger( ServiceDiscovery.class.getName());
	private final static Pattern pattern = compile( "(.*?)#{1}.*");

	protected final Properties properties = new Properties();
	protected final String system;

	/**
	 * Constructs a service information object without an associated system identifier.
	 */
	protected ServiceInfo() { this( null); }

	/**
	 * Constructs a service information object with the given system identifier. This
	 * identifier is used to query system specific JVM configuration files which allow
	 * a user or system administrator to override provider implementations.
	 * <p/>
	 * @see ServiceDiscovery
	 *
	 * @param system the identifier for the particular system, i.e. {@code net}, for
	 * which properties should be parsed.
	 */
	protected ServiceInfo( final String system) {
		this.system = system;
		if ( system == null) return;

		try {
			final File file = new File( System.getProperty( "java.home"), "lib/" + system + ".properties");
			if ( !file.exists()) logger.log( FINE, "No system settings for ''{0}''.", system);
			else {
				final FileInputStream input = new FileInputStream( file);
				properties.load( input);
				input.close();
			}
		} catch ( IOException e) {
			logger.log( WARNING, "Error reading properties.", e);
		}
	}

	/**
	 * Gets the system name, used to find system wide service properties.
	 *
	 * @return the system name, or {@code null} if none was set.
	 */
	public String getSystem() { return system; }

	/**
	 * Finds the first known implementation of the given service class (or interface) using the current
	 * class loader. This is equivalent to calling {@link #findService(ClassLoader,Class)} with {@code null}
	 * as the first parameter.
	 *
	 * @param service the service class (or interface) for which a compatible implementation will be returned.
	 * @return a concrete instance, or {@code null} if none is found.
	 */ 
	public < T> T findService( final Class< T> service) {
		return findService( null, service);
	}

	/**
	 * Finds the first known implementation of the given service class (or interface) using the provided
	 * class loader. This method searches, in order, a system property specified by the fully qualified
	 * service class name if one exists, the properties file located within the Java home directory
	 * if it exists, and finally the provided class loader by fetching all resources named according to
	 * the services specification. This method returns the first service implementation it finds.
	 *
	 * @param loader the class loader to use when searching for service descriptor files, or {@code null}
	 * to use the current class loader.
	 * @param service the service class (or interface) for which compatible implementations will be returned.
	 * @return a concrete instance, or {@code null} if none is found.
	 * @throws ClassCastException if a found provider class is not a subclass of {@code service}.
	 */ 
	public < T> T findService( final ClassLoader loader, final Class< T> service) {
		return doPrivileged( new PrivilegedAction< T>() {
			public T run() {
				T result = createInstance( loader, System.getProperty( service.getName()), service);
				if ( result != null) return result;

				result = createInstance( loader, properties.getProperty( service.getName()), service);
				if ( result != null) return result;

				final Collection< T> results = searchClassPath( loader, service);
				if ( !results.isEmpty()) return results.iterator().next();

				return null;
			}
		});
	}

	/**
	 * Finds all known implementations of the given service class using the current class loader.
	 * This is equivalent to calling {@link #findServices(ClassLoader,Class)} with {@code null} as the first parameter.
	 *
	 * @param service the service class (or interface) for which compatible implementations will be searched.
	 * @return a {@link Collection} of concrete instances, or the empty collection is none is found. 
	 * @throws ClassCastException if a found provider class is not a subclass of {@code service}.
	 */ 
	public < T> Iterable< T> findServices( final Class< T> service) {
		return findServices( null, service);
	}

	/**
	 * Finds all known implementations of the given service class using the provided class loader. This
	 * method searches, in order, a system property specified by the fully qualified service class name
	 * if one exists, the properties file located within the Java home directory if it
	 * exists, and finally the provided class loader by fetching all resources named according to the
	 * services specification. This method returns all compatible service implementations.
	 *
	 * @param loader the class loader to use when searching for service descriptor files, or {@code null}
	 * to use the current class loader.
	 * @param service the service class (or interface) for which compatible implementations will be searched.
	 * @return a {@link Collection} of concrete instances, or the empty collection is none is found. 
	 * @throws ClassCastException if a found provider class is not a subclass of {@code service}.
	 */ 
	public < T> Iterable< T> findServices( final ClassLoader loader, final Class< T> service) {
		return doPrivileged( new PrivilegedAction< Iterable< T>>() {
			public Iterable< T> run() {
				final Collection< T> results = new ArrayList< T>();
				T result = createInstance( loader, System.getProperty( service.getName()), service);
				if ( result != null) results.add( result);

				result = createInstance( loader, properties.getProperty( service.getName()), service);
				if ( result != null) results.add( result);

				results.addAll( searchClassPath( loader, service));
				return results;
			}
		});
	}

	/**
	 * Finds all service descriptors that contain an implementation of the given service. The resulting
	 * collection will contain an instance of each concrete provider implementation found through the
	 * given class loader, or the current class loader if {@code null} is specified.
	 *
	 * @param loader the class loader within which to search for service descriptors.
	 * @param service the service interface for which to find concrete implementations.
	 * @return a {@link Collection} of concrete instances of the service class (or interface), or an
	 * empty collection if none could be found.
	 * @throws ClassCastException if a found provider class is not a subclass of {@code service}.
	 */
	protected static < T> Collection< T> searchClassPath( ClassLoader loader, final Class< T> service) {
		final Collection< T> results = new ArrayList< T>();
		try {
			final String path = "META-INF/services/" + service.getName();
			logger.log( FINE, "Using path ''{0}''.", path);
			if ( loader == null) loader = ServiceDiscovery.class.getClassLoader();
			for ( URL url : list( loader.getResources( path))) {
				logger.log( FINE, "Reading url ''{0}''.", url);
				final BufferedReader reader = new BufferedReader( new InputStreamReader( url.openStream()));
				String line;
				while (( line = reader.readLine()) != null) {
					line = clean( line);
					final T result = createInstance( loader, line, service);
					logger.log( FINE, "Created instance ''{0}''.", result);
					if ( result != null) results.add( result);
				}
			}
		} catch ( IOException e) {
			assert true;
		}
		return results;
	}

	/**
	 * Attempts to create an instance of the named class using the provided class loader. If
	 * the given class loader is {@code null} then the current class loader is used. This
	 * method returns null if the class cannot be found or instantiated with a default constructor.
	 *
	 * @param loader the class loader to use to load the class definition, may be {@code null}.
	 * @param name name of the class to load and instantiate.
	 * @param service the service class of which the named class is expected to be a subclass.
	 * @return a new instance of the class indicated with {@code name}.
	 * @throws ClassCastException if the named class is not a subclass of {@code service}.
	 */
	protected static < T> T createInstance( final ClassLoader loader, final String name, final Class< T> service) {
		if ( name == null || name.length() == 0) return null;
		try {
			logger.log( FINE, "Creating instance of ''{0}'' as ''{1}'' using loader ''{2}''", new Object[] { name, service, loader});
			final Class< ?> clazz = loader == null ? forName( name) : loader.loadClass( name);
			return clazz.asSubclass( service).newInstance();
		} catch ( ClassNotFoundException e) {
			logger.log( WARNING, "Could not find the specified implementation class.", e);
			return null;
		} catch ( IllegalAccessException e) {
			logger.log( WARNING, "Could not access class definition.", e);
			return null;
		} catch ( InstantiationException e) {
			logger.log( WARNING, "The specified implementation is an interface or abstract class.", e);
			return null;
		}
	}

	/**
	 * Utility method to remove comments from service descriptors. This method removes any whitespace
	 * as well as comments beginning with {@code #} through the end of the line.
	 *
	 * @param line line including possible extra white space and comment.
	 * @return cleaned line or the empty string if the line is not valid.
	 */
	protected final static String clean( String line) {
		return pattern.matcher( line).replaceAll( "$1").trim();
	}
}
